      ////////////////////////////////////////
     // SPECSAISIE 1.2                      //
    // Written by Andrew Broad             ///
   //  BROADSOFT 2003                    ////
  //                                     /////
 // Memory.java (2002:11:09 17:21)      //////
//////////////////////////////////////////////
// Methods of Memory:                   //////
// - Memory constructor                 //////
// - poke                               //////
// - doke                               //////
// - peek                               //////
// - deek                               //////
// - loadSNA                            //////
// - saveSNA                            //////
// - loadSpectrumFile(SpectrumFile)     /////
// - loadSpectrumFile(SpectrumFile,int) ////
// - saveSpectrumBasicFile              ///
// - saveSpectrumCodeFile               //
/////////////////////////////////////////

import java.io.*;

/*****************************************************************************
** The class Memory represents the Spectrum memory and its contents.
*****************************************************************************/

public class Memory {
	public byte[] _memory;

	/********************************************************************/
	// Memory constructor : void ->
	/*********************************************************************
	** Construct an instance of Spectrum memory.
	*********************************************************************/

	public Memory () {
		_memory= new byte[65536];
	}

	/********************************************************************/
	// Memory.poke : int x byte -> void
	/*********************************************************************
	** Write a given byte value into a given memory location.
	** @param address The address of the memory location to poke.
	** @param value The byte value to write into the memory location.
	*********************************************************************/
	
	public void poke (int address, byte value) {
		if (! (address >= 0 && address <= 65535)) {
			System.err.println("Memory.poke: invalid address"
			+ address);
			System.exit(1);
		}
		_memory[address]= value;
	}

	/********************************************************************/
	// Memory.doke : int x int -> void
	/*********************************************************************
	** Write a given two-byte value into a given memory location.
	** @param address The address of the memory location to poke.
	** @param value The double-byte value to write into <TT>address</TT>
	**              and <TT>address+1</TT> location.
	*********************************************************************/
	
	public void doke (int address, int value) {
		if (! (address >= 0 && address <= 65535)) {
			System.err.println("Memory.doke: invalid address "
			+ address);
			System.exit(1);
		}
		if (! (value >= 0 && value <= 65535)) {
			System.err.println("Memory.doke: invalid value "
			+ value);
			System.exit(1);
		}
		_memory[address]= (byte)(value % 256);
		_memory[address+1]= (byte)(value / 256);
	}

	/********************************************************************/
	// Memory.peek : int -> byte
	/*********************************************************************
	** Reads the byte value of a specified memory location.
	** @param address The address of the memory location to peek.
	** @return The byte value stored in the memory location.
	*********************************************************************/
	
	public byte peek (int address) {
		if (! (address >= 0 && address <= 65535)) {
			System.err.println("Memory.peek - invalid address"
			+ address);
			System.exit(1);
		}
		return _memory[address];
	}

	/********************************************************************/
	// Memory.deek : int -> int
	/*********************************************************************
	** Reads the two-byte value of a specified memory location.
	** @param address The address of the memory location to peek.
	** @return The double-byte value stored in <TT>address</TT> and
	**         <TT>address+1</TT>.
	*********************************************************************/
	
	public int deek (int address) {
		if (! (address >= 0 && address <= 65535)) {
			System.err.println("Memory.deek - invalid address"
			+ address);
			System.exit(1);
		}
		return ByteSupport.twoByteValue(_memory[address],
		                                _memory[address+1]);
	}

	/********************************************************************/
	// Memory.loadSNA : String -> void
	/*********************************************************************
	** Load the memory from an SNA file.
	** @param filename The SNA file to load the memory from.
	*********************************************************************/

	public void loadSNA (String filename) {
		if (! filename.toUpperCase().endsWith(".SNA")) {
			System.err.println("Memory.loadSNA: was expecting a "
			+ "snapshot (.SNA) filename");
			System.exit(1);
		}
		FileInputStream snaFile= null;
		try {
			snaFile= new FileInputStream(filename);
		} catch (FileNotFoundException fnfe) {
			System.err.println("Memory.loadSNA: cannot open \""
			+ filename + "\"");
			System.exit(1);
		}
		try {
			int length= snaFile.available();
			if (length != 49179 && length != 131103) {
				System.err.println("Memory.loadSNA: bad leng"
				+ "th SNA file - should be 49179 bytes, not "
				+ length);
				System.exit(1);
			}
			if (length == 131103) {
				System.err.println("Warning: " + filename
				+ " is a 128K SNA file - SPECSAISIE only "
				+ "recognises the first 49179 bytes.\n");
			}
			snaFile.read(_memory,16357,49179);

		} catch (IOException ioe) {
			System.err.println("Memory.loadSNA: IOException");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Memory.saveSNA : String -> void
	/*********************************************************************
	** Save the memory as an SNA file.
	** @param filename The SNA file to save the memory as.
	*********************************************************************/

	public void saveSNA (String filename) {
		if (! filename.toUpperCase().endsWith(".SNA")) {
			System.err.println("Memory.saveSNA: was expecting a "
			+ "snapshot (.SNA) filename");
			System.exit(1);
		}
		try {
			FileOutputStream snaFile=
				new FileOutputStream(filename);

			snaFile.write(_memory,16357,49179);

			System.out.println(filename + " saved.");
		} catch (IOException ioe) {
			System.err.println("Memory.saveSNA: IOException");
			System.exit(1);
		}
	}

	/********************************************************************/
	// Memory.loadSpectrumFile : SpectrumFile -> void
	/*********************************************************************
	** Load a Spectrum file into memory. This is currently a very crude
	** approximation - it only works for CODE files, and does not set the
	** system variables as it should.
	** @param file The Spectrum file to load.
	*********************************************************************/

	public void loadSpectrumFile (SpectrumFile file) {
		for (int i= 0; i < file.getLength(); i++) {
			try {
				_memory[file.getParameter1() + i]
				= file._dataBlock[3 + i];
			} catch (ArrayIndexOutOfBoundsException aioobe) {
				System.err.println("ArrayIndexOutOfBoundsExce"
				+ "ption at Memory.loadSpectrumFile(" + file
				+ ")");
				System.err.println("_memory["
				+ (file.getParameter1() + i) + "]= "
				+ file + "._dataBlock[" + (3 + i) + "];");
				System.exit(1);
			}
		}
	}

	/********************************************************************/
	// Memory.loadSpectrumFile : SpectrumFile x int -> void
	/*********************************************************************
	** Load a Spectrum file into memory at a specified address instead of
	** the start address in the header. This is currently a very crude
	** approximation - it only works for CODE files, and does not set the
	** system variables as it should.
	** @param file The Spectrum file to load.
	** @param address The address to relocate to.
	*********************************************************************/

	public void loadSpectrumFile (SpectrumFile file, int address) {
		for (int i= 0; i < file.getLength(); i++) {
			try {
				_memory[address + i]= file._dataBlock[3 + i];
			} catch (ArrayIndexOutOfBoundsException aioobe) {
				System.err.println("ArrayIndexOutOfBoundsExce"
				+ "ption at Memory.loadSpectrumFile(" + file
				+ "," + address + ")");
				System.err.println("_memory[" + (address + i)
				+ "]= " + file + "._dataBlock[" + (3 + i)
				+ "];");
				System.exit(1);
			}
		}
	}

	/********************************************************************/
	// Memory.saveSpectrumBasicFile : String x int x int -> SpectrumFile
	/*********************************************************************
	** Saves a given interval as a Spectrum BASIC file of specified name.
	** @param filename A filename to save the Spectrum file as.
	** @param line The auto-start line-number (32768 if no auto-start).
	** @return The Spectrum file thus created.
	*********************************************************************/
	
	public SpectrumFile saveSpectrumBasicFile (String filename, int line)
	{
		int PROG= deek(23635);
		int varArea= deek(23627) - PROG;
		int length= deek(23641)-1 - PROG;
		SpectrumFile file= new SpectrumFile(filename, varArea, length,
		                                    line);
		for (int i= 0; i < length; i++) {
			file._dataBlock[3 + i]= _memory[PROG + i];
		}
		return file;
	}

	/********************************************************************/
	// Memory.saveSpectrumCodeFile : String x int x int -> SpectrumFile
	/*********************************************************************
	** Saves a given interval as a Spectrum CODE file of specified name.
	** @param filename A filename to save the Spectrum file as.
	** @param start The start address to save the Spectrum file from.
	** @param length The length (in bytes) of the Spectrum file to save.
	** @return The Spectrum file thus created.
	*********************************************************************/
	
	public SpectrumFile saveSpectrumCodeFile (String filename,
	                                          int start, int length)
	{
		SpectrumFile file= new SpectrumFile(filename, start, length);
		for (int i= 0; i < length; i++) {
			file._dataBlock[3 + i]= _memory[start + i];
		}
		return file;
	}
}
