      ///////////////////////////////////
     // SPECSAISIE 1.2                 //
    // Written by Andrew Broad        ///
   //  BROADSOFT 2003               ////
  //                                /////
 // Saisie.java (2002:10:26 20:49) //////
/////////////////////////////////////////
// Methods of Saisie:              /////
// - static saisie                 ////
// - static main                   ///
// - static processFlags           //
////////////////////////////////////

import java.io.*;

/*****************************************************************************
** Saisie is an application class for generating Spectrum BASIC code for
** typing in code - useful for typing code in emulator format into a real
** Spectrum.
*****************************************************************************/

public class Saisie {
	static boolean _includeREM= false;
	static String _REM= ": REM ";
	static boolean _outputToFiles= false;
	static PrintStream _out= System.out;
	static Memory _memory;

	/********************************************************************/
	// static Saisie.saisie : (int x int) -> void
	/*********************************************************************
	** Generates Spectrum BASIC code to type in a specified number of
	** bytes starting at the given start-address.
	** @param memory The representation of the Spectrum's memory.
	** @param start The address to start printing at.
	** @param end The address to stop printing at. However, it will always
	**            finish the line it's on, thus rounding up the number of
	**            bytes to the next multiple of eight.
	** @param includeREM If true, it plants a REM statement on the first
	**                   line, giving the file, start-address, and length.
	** @param filename The filename of the SNA file.
	** @return The Saisie output as a string (ending with a newline).
	*********************************************************************/

	public static String saisie (Memory memory, int start, int end,
	                             boolean includeREM, String filename)
	{
		String output= "";
		short line= 1000;
		for (int addr= start; addr <= end; addr+= 8) {
			output+= line + " DATA ";
			int checksum= (addr / 256) + (addr % 256);
			for (int i= 0; i < 8; i++) {
				int bite= ByteSupport.unsignByte
				          (memory._memory[addr + i]);
				output+= bite + ",";
				checksum+= (i+1) * bite;
			}
			checksum= 256 + (checksum % 744);
			output+= checksum;
			if (includeREM && addr == start) {
				output+= _REM + filename + " "
				         + start + "," + (end+1-start);
			}
			output+= "\n";
			line+= 10;
		}
		return output;
	}

	/********************************************************************/
	// static Saisie.main : String[] -> void
	/*********************************************************************
	** Application entry-point for generating a Spectrum BASIC program for
	** typing code into a real Spectrum. The program is generated as text
	** ready for printing out. Each line is a DATA statement with eight
	** bytes plus a checksum (256-999).
	** @param argv The vector of command-line arguments. There should be
	**             at least two arguments: an SNA file to load and a
	**             start-address. The optional third argument is the
	**             length of code to print (i.e. the number of bytes to
	**             type in), which defaults to 1024 if not specified. The
	**             optional fourth argument is the stop-address, for when
	**             you want to generate more than one type-in at once.
	*********************************************************************/

	public static void main (String argv[]) {
		String[] argv2= new String[argv.length];
		int argv2_length= 0;
		for (int i= 0; i < argv.length; i++) {
			if (argv[i].startsWith("-")) {
				processFlags(argv[i]);
			} else {
				argv2[argv2_length++]= argv[i];
			}
		}

		if (argv2_length < 2 || argv2_length > 4) {
			System.err.println("Usage: java Saisie <file.SNA> "
			+ "<start_address> (<length> (<stop_address>))");
			System.err.println("    Flags:-");
			System.err.println("-r: Include REM statement to show"
			+ " command-line arguments");
			System.err.println("-f: Write output to files rather "
			+ "than stdout");
			System.exit(1);
		}

		int startAddress= 32768;
		try {
			startAddress= Integer.parseInt(argv2[1]);
		} catch (NumberFormatException nfe) {
			System.err.println(argv2[1] + " doesn't parse as an "
			+ "integer!");
			System.exit(1);
		}

		int length= 1024;
		if (argv2_length >= 3) {
			try {
				length= Integer.parseInt(argv2[2]);
			} catch (NumberFormatException nfe) {
				System.err.println(argv2[2] + " doesn't parse"
				+ " as an integer!");
				System.exit(1);
			}
		}

		int stopAddress= startAddress;
		if (argv2_length >= 4) {
			try {
				stopAddress= Integer.parseInt(argv[3]);
			} catch (NumberFormatException nfe) {
				System.err.println(argv[3] + " doesn't parse "
				+ "as an integer!");
				System.exit(1);
			}
		}

		_memory= new Memory();
		_memory.loadSNA(argv2[0]);

		for (int address= startAddress; address <= stopAddress;
		                                address+= length)
		{
			if (_outputToFiles) {
				try {
					_out= new PrintStream
					      (new FileOutputStream
			 		      (String.valueOf(address)));
				} catch (FileNotFoundException fnfe) {
					System.err.println(fnfe);
					System.exit(1);
				}
			}
			_out.print(saisie(_memory, address, address+length-1,
			                  _includeREM, argv2[0]));
		}
	}

	/********************************************************************/
	// static Saisie.processFlags : String[] -> void
	/*********************************************************************
	** A subroutine to process the flags in a command-line argument
	** starting with `-'. This method assumes that it is only called with
	** strings starting with `-', and ignores the first character.
	** @param arg A `flags' command-line argument (i.e. `-' followed by
	**            letters).
	**            -r: Include REM statement to show the command-line
	**                arguments in first line of output.
	**            -f: Write output to files rather than System.out.
	*********************************************************************/

	private static void processFlags (String arg) {
		char[] flags= arg.toCharArray();

		for (int i= 1; i < flags.length; i++) {
			switch (flags[i]) {
				case 'r':
					_includeREM= true;
					break;
				case 'f':
					_outputToFiles= true;
					break;
				default:
					System.err.println("Warning: -"
					+ flags[i] + " not recognised.");
			}
		}
	}
}