      /////////////////////////////////////////
     // SPECSAISIE 1.2                       //
    // Written by Andrew Broad              ///
   //  BROADSOFT 2003                     ////
  //                                      /////
 // SpectrumFile.java (2002:11:09 17:09) //////
///////////////////////////////////////////////
// Methods of SpectrumFile:              //////
// - SpectrumFile constructors           //////
// - readHeader                          //////
// - readDataBlock                       //////
// - readNextByte                        //////
// - write                               //////
// - writeHeader                         //////
// - writeDataBlock                      //////
// - writeByte                           //////
// - getType                             //////
// - getName                             /////
// - getLength                           ////
// - getParameter1                       ///
// - getParameter2                       //
//////////////////////////////////////////

import java.io.*;

/*****************************************************************************
** The class SpectrumFile represents a file on a Spectrum tape - its header
** and its data block.
*****************************************************************************/

//////////////////////////////////////////////////////////////////////////////
// Structure of header:                                                     //
// - Offsets  0-1  : How many bytes will follow (always 19)                 //
// - Offset   2    : flag byte (always 0)                                   //
// - Offset   3    : file type (0=BASIC, 1=NUM.ARRAY, 2=CHAR.ARRAY, 3=CODE) //
// - Offsets  4-13 : filename                                               //
// - Offsets 14-15 : length of data block                                   //
// - Offsets 16-17 : parameter 1 } file-type                                //
// - Offsets 18-19 : parameter 2 } specific                                 //
// - Offset  20    : header checksum byte                                   //
//                                                                          //
// Structure of data block:                                                 //
// - Offsets  0-1  : How many bytes will follow                             //
// - Offset   2    : flag byte (non-zero)                                   //
// - Offsets  3-n  : the data                                               //
// - Offset  (n+1) : data checksum byte                                     //
//                                                                          //
// [http://www.srcf.ucam.org/~pak21/cssfaq/fileform.html]                   //
//////////////////////////////////////////////////////////////////////////////

public class SpectrumFile {
	public static final byte BASIC= 0;
	public static final byte NUMBER_ARRAY= 1;	
	public static final byte CHARACTER_ARRAY= 2;
	public static final byte CODE= 3;

	public byte[] _header;
	public byte[] _dataBlock;

	/********************************************************************/
	// SpectrumFile constructor : FileInputStream ->
	/*********************************************************************
	** Construct a Spectrum file by reading it from a TAP file.
	** This constructor should not be used directly - it is called from
	** within <A HREF="Tape.html">Tape</A>.
	** @param fis A file input stream to read the Spectrum file from,
	**            assumed to be wound to the start of a file.
	*********************************************************************/

	public SpectrumFile (FileInputStream fis) {
		readHeader(fis);
		readDataBlock(fis);
	}

	/********************************************************************/
	// SpectrumFile constructor : String x int x int x int ->
	/*********************************************************************
	** Construct a Spectrum BASIC file with specified filename and
	** auto-start line-number.
	** @param name The filename for the Spectrum file.
	** @param varArea Offset for the address of the variable-area.
	** @param length The length.
	** @param line The auto-start line-number (32768 if no auto-start).
	*********************************************************************/

	public SpectrumFile (String name, int varArea, int length, int line) {
		_header= new byte[20];
		_header[0]= 19; _header[1]= 0; // 19 bytes will follow
		_header[2]= 0;                 // flag byte
		_header[3]= 0;                 // BASIC file
		for (int n= 4; n <= 13; n++) { // filename
			try {
				_header[n]= (byte)name.charAt(n-4);
			} catch (StringIndexOutOfBoundsException sioobe) {
				_header[n]= 32;
			}
		}
		_header[14]= (byte)(length%256);
		_header[15]= (byte)(length/256);
		_header[16]= (byte)(line%256);
		_header[17]= (byte)(line/256);
		_header[18]= (byte)(varArea%256);
		_header[19]= (byte)(varArea/256);

		_dataBlock= new byte[length+3];
		_dataBlock[0]= (byte)((length+2)%256);
		_dataBlock[1]= (byte)((length+2)/256);
		_dataBlock[2]= -1; // flag byte (non-zero)
	}

	/********************************************************************/
	// SpectrumFile constructor : String x int x int ->
	/*********************************************************************
	** Construct a Spectrum CODE file with specified filename,
	** start address and length.
	** @param name The filename for the Spectrum file.
	** @param start The start-address.
	** @param length The length of the file.
	*********************************************************************/

	public SpectrumFile (String name, int start, int length) {
		_header= new byte[20];
		_header[0]= 19; _header[1]= 0; // 19 bytes will follow
		_header[2]= 0;                 // flag byte
		_header[3]= 3;                 // CODE file
		for (int n= 4; n <= 13; n++) { // filename
			try {
				_header[n]= (byte)name.charAt(n-4);
			} catch (StringIndexOutOfBoundsException sioobe) {
				_header[n]= 32;
			}
		}
		_header[14]= (byte)(length%256);
		_header[15]= (byte)(length/256);
		_header[16]= (byte)(start%256);
		_header[17]= (byte)(start/256);
		_header[18]= 0; _header[19]= -128;

		_dataBlock= new byte[length+3];
		_dataBlock[0]= (byte)((length+2)%256);
		_dataBlock[1]= (byte)((length+2)/256);
		_dataBlock[2]= -1; // flag byte (non-zero)
	}

	/********************************************************************/
	// SpectrumFile.readHeader : FileInputStream -> void
	/*********************************************************************
	** Read the header of the next file on the tape.
	** @param fis A file input stream representing the tape.
	*********************************************************************/

	private void readHeader (FileInputStream fis) {
		_header= new byte[20];

		_header[0]= readNextByte(fis);
		_header[1]= readNextByte(fis);
		byte checksum= _header[2]= readNextByte(fis);
		for (int i= 3; i < 20; i++) {
			checksum^= _header[i]= readNextByte(fis);
		}
		if (checksum != readNextByte(fis)) {
			System.err.println("WARNING: Checksum error in "
			+ "header!");
		}
		printHeader();
	}

	/********************************************************************/
	// SpectrumFile.readDataBlock : FileInputStream -> void
	/*********************************************************************
	** Read the data block of this file on the tape.
	** @param fis A file input stream representing the tape, which is
	**            expected to be wound to the start of the data block
	**            (i.e. after the header).
	*********************************************************************/

	private void readDataBlock (FileInputStream fis) {
		byte l= readNextByte(fis);
		byte h= readNextByte(fis);
		_dataBlock= new byte[ByteSupport.twoByteValue(l,h) + 1];
		_dataBlock[0]= l;
		_dataBlock[1]= h;

		byte checksum= _dataBlock[2]= readNextByte(fis);
		for (int i= 3; i < _dataBlock.length; i++) {
			checksum^= _dataBlock[i]= readNextByte(fis);
		}
		if (checksum != readNextByte(fis)) {
			System.err.println("WARNING: Checksum error in "
			+ "data block!");
		}
		System.out.println();
	}

	/********************************************************************/
	// SpectrumFile.readNextByte : FileInputStream -> byte
	/*********************************************************************
	** Reads and returns the next byte from the TAP file. This will be
	** interpreted as a SIGNED 8-bit integer - to avoid the overhead of
	** converting every byte to an unsigned 8-bit integer, it is the
	** caller's responsibility to do so when necessary.
	** @param fis The file input stream representing the TAP file.
	** @return The value of the two bytes as an unsigned 16-bit integer.
	*********************************************************************/

	private byte readNextByte (FileInputStream fis) {
		try {
			if (fis.available() > 0) {
				byte b= (byte)fis.read();
				return b;
			}
			else {
				System.err.println("SpectrumFile.readNextByte"
				+ " failed because no bytes were available to"
				+ " read.");
				System.exit(1); return 0;
			}
		} catch (IOException ioe) {
			System.err.println("IOException thrown in " +
			                   "SpectrumFile.readNextByte");
			System.exit(1); return 0;
		}
	}

	/********************************************************************/
	// SpectrumFile.write : FileOutputStream -> void
	/*********************************************************************
	** Write this Spectrum file to a TAP file. This method should not be
	** invoked directly - it is called from within
	** <A HREF="Tape.html">Tape</A>.
	** @param fos A file output stream to write the Spectrum file to.
	*********************************************************************/

	public void write (FileOutputStream fos) {
		writeHeader(fos);
		writeDataBlock(fos);
	}

	/********************************************************************/
	// SpectrumFile.writeHeader : FileOutputStream -> void
	/*********************************************************************
	** Write the header of this Spectrum file to the TAP file.
	** @param fis A file output stream representing the tape.
	*********************************************************************/

	private void writeHeader (FileOutputStream fos) {
		writeByte(_header[0], fos);
		writeByte(_header[1], fos);
		byte checksum; writeByte(checksum= _header[2], fos);
		for (int i= 3; i < 20; i++) {
			writeByte(_header[i],fos);
			checksum^= _header[i];
		}
		writeByte(checksum, fos);
	}

	/********************************************************************/
	// SpectrumFile.writeDataBlock : FileOutputStream -> void
	/*********************************************************************
	** Write the data block of this Spectrum file to the TAP file.
	** @param fis A file output stream representing the tape, which is
	**            expected to be at the start of the data block
	**            (i.e. after the header).
	*********************************************************************/

	private void writeDataBlock (FileOutputStream fos) {
		writeByte(_dataBlock[0], fos);
		writeByte(_dataBlock[1], fos);
		byte checksum; writeByte(checksum= _dataBlock[2], fos);
		for (int i= 3; i < _dataBlock.length; i++) {
			writeByte(_dataBlock[i], fos);
			checksum^= _dataBlock[i];
		}
		writeByte(checksum, fos);
	}

	/********************************************************************/
	// SpectrumFile.writeByte : (byte x FileOutputStream) -> void
	/*********************************************************************
	** Writes a byte to the TAP file.
	** @param b The byte to write.
	** @param fos The file output stream representing the TAP file.
	** @return The value of the two bytes as an unsigned 16-bit integer.
	*********************************************************************/

	private void writeByte (byte b, FileOutputStream fos) {
		try {
			fos.write(b);
		} catch (IOException ioe) {
			System.err.println("IOException thrown in " +
			                   "SpectrumFile.writeByte");
			System.exit(1);
		}
	}

	/********************************************************************/
	// SpectrumFile.getType : void -> byte
	/*********************************************************************
	** Returns the type of this Spectrum file.
	** @return The type of this Spectrum file (0 = BASIC, 1 = NUMBER
	**         ARRAY, 2 = CHARACTER ARRAY, 3 = CODE).
	*********************************************************************/

	public byte getType () {
		return _header[3];
	}

	/********************************************************************/
	// SpectrumFile.getName : void -> String
	/*********************************************************************
	** Returns the name of this Spectrum file.
	** @return The name of this Spectrum file.
	*********************************************************************/

	public String getName () {
		String name= "";
		for (int n= 4; n <= 13; n++) {
			name+= (char)_header[n];
		}
		return name;
	}

	/********************************************************************/
	// SpectrumFile.getLength : void -> int
	/*********************************************************************
	** Returns the number of bytes of data in this Spectrum file.
	** @return The number of bytes of data in this Spectrum file.
	*********************************************************************/

	public int getLength () {
		return ByteSupport.twoByteValue(_header[14],_header[15]);
	}

	/********************************************************************/
	// SpectrumFile.getParameter1 : void -> int
	/*********************************************************************
	** Returns Parameter 1 of this Spectrum file. For a BASIC file,
	** Parameter 1 is the autostart line number, or >= 32768 if there
	** isn't one. For a CODE file, Parameter 1 is the start address.
	** @return Parameter 1 of this Spectrum file.
	*********************************************************************/

	public int getParameter1 () {
		return ByteSupport.twoByteValue(_header[16],_header[17]);
	}

	/********************************************************************/
	// SpectrumFile.getParameter2 : void -> int
	/*********************************************************************
	** Returns Parameter 2 of this Spectrum file. For a BASIC file,
	** Parameter 2 is the start of the variable area relative to the start
	** of the program. For a CODE file, Parameter 2 should be 32768.
	** @return Parameter 2 of this Spectrum file.
	*********************************************************************/

	public int getParameter2 () {
		return ByteSupport.twoByteValue(_header[18],_header[19]);
	}

	/********************************************************************/
	// SpectrumFile.printHeader : void -> void
	/*********************************************************************
	** Prints the header of this Spectrum file in the standard
	** "Filetype: filename" format.
	*********************************************************************/

	public void printHeader () {
		switch (getType()) {
			case 0:
				System.out.print("Program: ");
				break;
			case 1:
				System.out.print("Number array: ");
				break;
			case 2:
				System.out.print("Character array: ");
				break;
			case 3:
				System.out.print("Bytes: ");
				break;
		}
		System.out.println(getName());
	}
}
