      /////////////////////////////////
     // SPECSAISIE                   //
    // Written by Andrew Broad      ///
   //  BROADSOFT 2000             ////
  //                              /////
 // Tape.java (2000:03:09 13:24) //////
///////////////////////////////////////
// Methods of Tape:              //////
// - Tape constructor ()         //////
// - Tape constructor (String)   //////
// - loadTAP                     //////
// - saveTAP                     //////
// - getName                     /////
// - getFiles                    ////
// - addFile                     ///
// - findFile                    //
//////////////////////////////////

import java.io.*;
import java.util.*;

/*****************************************************************************
** The class Tape is an internal representation of a TAP file.
*****************************************************************************/

public class Tape {
	String _name;
	Vector/*SpectrumFile*/ _files= new Vector();

	/********************************************************************/
	// Tape constructor : void ->
	/*********************************************************************
	** Construct a tape. The callee should subsequently call either
	** loadTAP to load a TAP file into the tape object, or saveTAP to
	** save it as a TAP file.
	*********************************************************************/

	public Tape () {}

	/********************************************************************/
	// Tape constructor : String ->
	/*********************************************************************
	** Construct a tape and load it from a TAP file.
	** @param tapeName A filename ending in .TAP for this tape.
	*********************************************************************/

	public Tape (String tapeName) {
		loadTAP(tapeName);
	}

	/********************************************************************/
	// Tape.loadTAP : String -> void
	/*********************************************************************
	** Load the TAP file (whose name was specified in the constructor)
	** into this tape object. For tapes that are to be loaded from TAP
	** files (as opposed to created internally), the callee should invoke
	** this method immediately after the constructor.
	** @param filename A filename ending in .TAP for this tape.
	*********************************************************************/

	public void loadTAP (String filename) {
		if (! filename.toUpperCase().endsWith(".TAP")) {
			System.err.println("Tape.loadTAP: was expecting a "
			+ "tape (.TAP) filename");
			System.exit(1);
		}
		_name= filename;

		FileInputStream fis= null;
		try {
			fis= new FileInputStream(_name);
		} catch (FileNotFoundException fnfe) {
			System.err.println("Tape.loadTAP: cannot open\""
			+ _name + "\"");
			System.exit(1);
		}
		////////////////
		// Read files //
		////////////////
		System.out.println("--------------------------------\n");
		try {
			while (fis.available() > 0) {
				addFile(new SpectrumFile(fis));
			}
		}
		catch (IOException ioe) {
			System.err.println("IOException caught in Tape."
			+ "loadTAP(\"" + _name + "\"");
			System.exit(1);
		}
		System.out.println("--------------------------------");
	}

	/********************************************************************/
	// Tape.saveTAP : String -> void
	/*********************************************************************
	** Save this tape as a TAP file.
	** @param filename A filename ending in .TAP for this tape.
	*********************************************************************/

	public void saveTAP (String filename) {
		if (! filename.toUpperCase().endsWith(".TAP")) {
			System.err.println("Tape.saveTAP: was expecting a "
			+ "tape (.TAP) filename");
			System.exit(1);
		}
		_name= filename;

		FileOutputStream fos= null;
		try {
			fos= new FileOutputStream(_name);
		} catch (FileNotFoundException fnfe) {
			System.err.println("Tape.saveTAP: cannot open\""
			+ _name + "\"");
			System.exit(1);
		}
		/////////////////
		// Write files //
		/////////////////
		Enumeration files= _files.elements();
		while (files.hasMoreElements()) {
			((SpectrumFile)files.nextElement()).write(fos);
		}
	}

	/********************************************************************/
	// Tape.getName : void -> String
	/*********************************************************************
	** Returns the name of this tape (the filename of the TAP file).
	** @return The name of this tape.
	*********************************************************************/

	public String getName () {
		return _name;
	}

	/********************************************************************/
	// Tape.getFiles : void -> Vector<SpectrumFile>
	/*********************************************************************
	** Returns the vector of Spectrum files on this tape.
	** @return The vector of Spectrum files on this tape (each element is
	**         an object of type
	**         <A HREF="SpectrumFile.html">SpectrumFile</A>).
	*********************************************************************/

	public Vector getFiles () {
		return _files;
	}

	/********************************************************************/
	// Tape.addFile : SpectrumFile -> void
	/*********************************************************************
	** Adds a Spectrum file to the tape (at the end).
	** @param A Spectrum file.
	*********************************************************************/

	public void addFile (SpectrumFile file) {
		_files.addElement(file);
	}

	/********************************************************************/
	// Tape.findFile : String x int -> SpectrumFile
	/*********************************************************************
	** Find a file of a given name and type on the tape.
	*********************************************************************/

	public SpectrumFile findFile (String filename, int filetype) {
		Enumeration files= _files.elements();
		while (files.hasMoreElements()) {
			SpectrumFile file= (SpectrumFile)files.nextElement();
			if (file.getName() == filename
			 && file.getType() == filetype) {
				return file;
			}
		}
		System.err.println("Tape.findFile: File \"" + filename
		+ "\" of type " + filetype + " not found.");
		System.exit(1); return null;
	}
}
